#!/usr/bin/perl
# -*- perl -*-
eval 'exec /usr/bin/perl -S $0 ${1+"$@"}'
    if 0;

## use IPC::Run qw/start run finish/;
use strict;
use Carp;
use Memoize;
use UNIVERSAL qw(isa can);

use TAG;
use TAG::LP;

use AppConfig qw/:argcount :expand/;

use IPC::Run qw/start run/;
use IPC::Open2;

use POSIX qw(locale_h);
use locale;
setlocale(LC_ALL,"fr");

use Data::Dumper;
use IO::Handle;

use File::Temp qw/tempfile/;

use lib '/usr/share/frmg';
use MyRegExp;
use LexedGram;

use XML::Twig;

use Clone;

## Needed for user nobody to be able to load Data::Compare from some place where it has enough rights
BEGIN { chdir '/tmp' };

use Data::Compare;

use Lingua::Features;

my $config = AppConfig->new(
			    "lexed|l=f"     => {DEFAULT => "/usr/bin/lexed"},
			    "dicodir|p=f"   => {DEFAULT => "/usr/lib/lexique" },
			    "datadir|p=f"   => {DEFAULT => "/usr/share/frmg" },
			    "tmpdir=f"      => {DEFAULT => "/tmp"},
			    "dico|d=s"      => {DEFAULT => 'dico.xlfg' },
			    "macros|m=f"    => {DEFAULT => "dico.xlfg.templates"},
			    "locutions=f"   => {DEFAULT => "dico.xlfg.cmpnd"},
			    "prefixes=f"    => {DEFAULT => "dico.sxspell.pref"},
			    "map=f"         => {DEFAULT => "frgram.map"},
			    "complete=f"    => {DEFAULT => "complete.lex" },
			    "expand!"       => {DEFAULT => 1},
			    "vcomp_expand!" => {DEFAULT => 1},
			    "sxpipe!"       => {DEFAULT => 1},
			    "sxpipecmd=f"   => {DEFAULT => '/usr/local/bin/sxpipe'},
			    "comment!"      => {DEFAULT => 0},
			    "cache|c=f",
			   );

my $conffile;

foreach ("$ENV{HOME}/.frmgrc", "/etc/frmg.conf") {
  $conffile=$_, last if -f $_ && -r $_;
}

# read configuration file
if ($conffile) {
  $config->file("$conffile")
    || die "can't open or process configuration file $conffile";
}

$config->args();

my $datadir = $config->datadir();
my $dicodir = $config->dicodir();
my $dico = $config->dico();
my $macrosfile = $dicodir . '/' . $config->macros();
my $locutionsfile = $dicodir . '/' . $config->locutions();
my $prefixesfile = $dicodir . '/' . $config->prefixes();
my $mapfile = $datadir . '/' . $config->map();
my $completefile = $datadir . '/' . $config->complete();
my $expand = $config->expand();
my %unknown = ();

my %cachelexed = ();

my $map = $TAG::LP::map = {};

parse_map_file($map, $mapfile);

TAG::LP::header_reader($datadir . '/' . $map->{'lp'}{'header'}) if (defined $map->{'lp'}{'header'});

memoize 'macro_expand';
##memoize 'entry_parse';

my @lexed = ($config->lexed,'-d',"$dicodir",'-p',"$dico",'consult');

## print "CMD @lexed\n";

my %locutions = ();

if (-r $locutionsfile) {
  open LOC,"<",$locutionsfile
    ||  die "can't open locution file: $!";
  while (<LOC>) {
    chomp;
    my @seq = split(/\s+/,$_);
    while (@seq) {
      my $prefix = join(' ',@seq);
      $prefix =~ s/' /'/og;
      $locutions{$prefix} ||= 1;
      pop @seq;
    }
  }
  close LOC;
}

my %prefixes = ();

if (-r $prefixesfile) {
  open PREF,"<",$prefixesfile
    ||  die "can't open prefixes file: $!";
  while (<PREF>) {
    next if m{^//};
    next if /^\s*$/;
    chomp;
    my ($pref) = split(/\s+/,$_);
    $prefixes{$pref} = 1;
  }
  close PREF;
}

## Hash of Hash of Feature Structures lemme->cat->fs
my $complete = {};

my $prefpat = join('|',sort {length($b) <=> length($a)} keys %prefixes);
$prefpat = qr/$prefpat/;

my %agglutinates = ( "au" => ":prep le__det",
		     "aux" => ":prep les__det",
		     "auquel" => ":prep lequel",
		     "auxquels" => ":prep lesquels",
		     "auxquelles" => ":prep lesquelles",
		     "du" => "de:prep le__det",
		     "duquel" => "de:prep lequel",
		     "des" => "de:prep les__det",
		     "desquels" => "de:prep lesquels",
		     "desquelles" => "de:prep lesquelles",
		     "s" => "en:prep les__det",
		     "audit" => ":prep ledit",
		     "auxdits" => ":prep lesdits",
		     "auxdites" => ":prep lesdites",
		     "dudit" => "de:prep ledit",
		     "desdits" => "de:prep lesdits",
		     "desdites" => "de:prep lesdites",
		    );
my $agglpat = join('|',keys %agglutinates);
$agglpat = qr/$agglpat/i;

my $lexinfo = qr/\{\{(.*?)\}\}\s*/o;

my %macro = ();

my %allowed_cat = ( 'n' => 1,
		    'v' => 1,
		    'cl' => 1,
		    'd' => 1,
		    'adj' => 1
		  );

my %cat_transfer = (
		    'auxEtre' => 'aux',
		    'auxAvoir' => 'aux'
		   );

sub normalize_cat {
  my $cat = shift;
  return $cat_transfer{$cat} || $cat;
}


my %lingua = ();

foreach my $type (Lingua::Features::FeatureType->types()) {
  foreach my $idv ($type->values) {
    ##    print "Register type=".$type->id." idv=$idv v=".$type->value_name($idv)."\n";
    $lingua{$type->value_name($idv)} = $idv;
  }
}


my %value_transfer = (
		      'ssubj' => 'subj',
		      'vsubj' => 'subj',
		      'intens' => 'intensive',

		      # Should use either short or long normalized values provided by Lingua::Features
		      # but unfortunately, we mix both notation for the grammar !
		      # => have to establish a manual list
		      'singular' => 'sg',
		      'plural' => 'pl',
		      'sing' => 'sg',
		      'true' => '+',
		      'false' => '-',
		     );

sub normalize_value {
  my $v = shift;
##  return $lingua{$v} || $value_transfer{$v} || $v;
  return $value_transfer{$v} || $v;
}

my %exclude_feature = (
##		       'lightverb' => 1,
		       'loc' => 1, # to be activated as soon as possible
		       'form' => 1,
		       'persposs' => 1
		      );

my %feature_transfer = (
			'v-form' => 'mode',
			'form-aux' => 'form_aux',
			'aux-req' => 'aux_req',
			'demonstrative' => 'dem',
			'define' => 'def',
			'locative' => 'loc',
			'possessive' => 'poss',
			'inflnumber' => 'number',
			'inflperson' => 'person',
			'qu' => 'wh',
			## following ones are to reverse from English to French !
			'mood' => 'mode'
		       );


my %lingua_feature_transfer = (
			       'mood' => 'mode',
			       'number' => 'num',
			       'person' => 'pers',
			       'numberposs' => 'num',
			       'persposs' => 'pers',
			       'poss' => 'bool'
);


## List of features whose values are used in long form in the grammar
my %lingua_value_expand = (
			   'tense' => 1,
			   'mode' => 1,
			   'mood' => 1
);

sub normalize_feature {
  my $f = shift;
  return $feature_transfer{$f} || $f;
}

my %pcas_transfer = (
		    );

sub normalize_pcas {
  my $p = shift;
  return $pcas_transfer{$p} || $p;
}

## use NumbersGram;
## my $nbparser = new NumbersGram();

#$::RD_TRACE=1;

#$::RD_AUTOACTION = q{ $item[1] };

#$::RD_HINT=1;

sub features {
  my @l = @_;
  my @fs;
  while (@l) {
    my $f = shift @l;
    my $v = shift @l;
    next if $exclude_feature{$f};
    push(@fs,$f => TAG::Comp->new( {name => $f,
				    rel => 'eq',
				    value => $v
				   }
				 ));
  }
  return @fs;
}

sub val {
  my $v = shift;
  return TAG::Val->new({text=>$v});
}

sub syntax {
  my $parser = shift;
  my $lemma = shift;
  my $subcat = shift;
  my $pos = 0;
  my $refl = ($lemma =~ s/Se$//o) ? TAG::Plus->new() : TAG::Minus->new() ;
  my %args = ();
  my @subcat = @$subcat;
#  print "SYNTAX $lemma $subcat\n";
  ## print Data::Dumper->Dump([$subcat],['subcat']);
  foreach my $arg ('arg0' .. 'arg2') {
    $args{$arg} = subcat (shift @subcat || [ TAG::Minus->new() ]);
  }
  # print "SUBCAT ",dump(@args),"\n";
  my @args = features( lemma =>  [ val($lemma) ],
		       refl => [$refl],
		       %args
		     );
  my $ht = TAG::Feature->new({type=>'ht',f=> {@args }});
##  print "SUBCAT ",Data::Dumper->Dump([$ht],['ht']),"\n";
  return $ht;
}

sub subcat {
  my $vals = shift;
  my @pcas = ();
  my @vals = ();
  my %kind = ();
  if ($config->vcomp_expand()
      && grep( $_->{text}  && $_->{text} eq 'scomp', @$vals) 
      && !grep( $_->{text} && $_->{text} eq 'vcomp', @$vals) ) {
    ## to overcome some current default of the lexicon
    push(@$vals,val('vcomp'),val('-vcomp'),val('de-vcomp'));
  }
  my %pcas = ();
  foreach my $val (@$vals) {
##    next unless (defined $val->{text});
    $val->{text} ||= "";
    next if ($val->{text} && $val->{text} =~ /^arg\d/);
    my $newval = $val;
    if ($val->isa('TAG::Val')) {
      if  ($val->{text} && $val->{text} =~ /^(\S+)-(\S+)/) {
	my $pcas = normalize_pcas($1);
	push(@pcas,val($pcas)) unless (exists $pcas{$pcas});
	$pcas{$pcas} ||= 1;
##	next if (@pcas > 1);	# no need to push another prepobj if already present
	$newval = val("prep$2");
      }
    }
    next if ($newval->{text} && exists $kind{$newval->{text}});
    $kind{$newval->{text}} = 1;
    push(@vals,$newval);
  }
  push(@pcas, TAG::Minus->new());
  return  [ 
	   TAG::Feature->new( { type => 'arg',
				f => { features( 'kind' => [@vals],
						 'pcas' => [@pcas]
					       )
				     }
			      } ) 
	  ];
}

sub entry {
  my $parser = shift;
  my $cat = shift;
  my $fs = shift;
  $fs->{type} = $cat;
  return TAG::LemmaRef->new({ cat => $cat, feature => $fs });
}

sub TAG::LemmaRef::lexLP {
  my $self = shift;
  my $fs2 = shift;
  my $name = quote($self->{'name'});
  my $cat = quote($self->{'cat'});
  my $lemma = '_';
  my $feature = $self->{'feature'};
  my $ht;
  my $xlemma = $self->{name};
  my $xcat = $self->{cat};
  if ($feature) {
    
    if (defined $fs2) {
      return unless (fs_unify($feature,$fs2));
    }

    ## Special keys
    delete $feature->{f}{cat};

    my @ctrkeys = grep(/^(\S+-)?(ssubj|subj|obj|acomp|vcomp|scomp)/,keys %{$feature->{f}});

    foreach my $key (@ctrkeys) {
    }

    delete $feature->{f}{$_} foreach @ctrkeys;

    if(defined $feature->{f}{pred}) {
      $ht = $feature->{f}{pred}{value}->[0];
##      print Data::Dumper->Dump([$ht->{f}{lemma}{value}->[0]],['ht']);
      $xlemma = $ht->{f}{lemma}{value}->[0]{'text'};
      $xlemma =~ s/_+(\d+)$//og; # normalizing lemma
      $lemma = quote($xlemma);
##      print "LEMMA $lemma\n";
      delete $ht->{f}{lemma};
      $ht = $ht->LP('ht');
      delete $feature->{f}{pred};
    }
    
    if (defined $feature->{f}{fct} 
	## need to check that expansion has not yet been done
	## because of interactions with memoization
	&& $feature->{f}{fct}{value}[0]->isa('TAG::Val') 
       ) {
      my $subcat = $feature->{f}{fct}{value};
#      print "FCT ",Data::Dumper->Dump([$subcat],['subcat']),"\n";
      $feature->{f}{fct} = TAG::Comp->new( {name => 'fct',
					    rel => 'eq',
					    value => subcat($subcat)
					    } );
#      print "FCT2 ",Data::Dumper->Dump([$subcat],['subcat']),"\n";
    }

    # do not handle support verbs (lightverbs) at this time
##    return  if (exists $feature->{f}{lightverb});


    if ($complete->{$lemma}{$cat}) {
##      print "%% Found completion entry\n";
      my $cfs = $complete->{$lemma}{$cat};
      ## Perform very limited unification by adding features not already present
      foreach my $key (keys %{$cfs->{f}}) {
	$feature->{f}{$key} = $cfs->{f}{$key} unless (exists $feature->{f}{$key});
      }
    }

    $feature = $feature->LP($cat);
  } else {
    $feature = '_';
  }
  $ht = '_' unless defined $ht;
  return { lemma => $lemma, 
	   cat => $cat, 
	   top => $feature,
	   ht => $ht,
	   xlemma => $xlemma
	 }
}

sub entry_print {
    my ($self,$lex,$left,$right,$nlex) = @_;
    $lex = quote($lex);
    $nlex =~ s/qu'$/que/o;	# Should be properly handled by the lexicon !
    $nlex = quote($nlex);
    my $lemma = ($self->{lemma} eq '_') ? $nlex : $self->{lemma};
##    $lemma =~ s/_+(\d+)$/_$1/og; # if we wish to keep a trace of which lemma
    print <<EOF
'C'($left, 
    lemma{ lex => $nlex,
           truelex => $lex,
	   lemma => $lemma,
	   cat => $self->{cat},
	   top => $self->{top},
	   anchor  => tag_anchor{ name => $self->{ht}, coanchors => [], equations => [] }
	 },
    $right
   ).

EOF
}

my $parser = new LexedGram();

read_macrosfile($macrosfile);

## Reading extra info to complete some entries
if (-r $completefile) {
  open COMP,"<",$completefile 
    ||  die "can't open complete file: $!";
  while (<COMP>) {
    next if m{^//};
    next if /^\s*$/;
    next if /^#/;
    chomp;
    my ($lemma,$cat,$fs) = split(/\t+/,$_,3);
##    print "Handling complete '$lemma' '$cat' '$fs'\n";
    $fs = $parser->fs($fs);
    $complete->{$lemma}{$cat} = $fs;
##    print "Register complete info for $lemma $cat\n";
  }
  close COMP;
}


my @fifo = ();
my @pending = ();
my $nbsentence = 0;

my $lexin = '';
my $lexout = '';
my $h = start \@lexed,\$lexin,\$lexout;

my $dagmode = 0;
my $dagmax = 0;

my %mafstruct = ();
my $mafcurrent = 0;


$|=1;

if ($config->sxpipe) {
  my $sxpipecmd = $config->sxpipecmd();
  my $cmd = "$sxpipecmd | /usr/local/bin/unfold_dag.out";
  while (<>) {
    if (/^##\s*DAG\s+BEGIN\s*(.*)/oi) {
      line_handler(\$_);
      while (<>) {
	line_handler(\$_);
	last if (/^##\s*DAG\s+END/oi);
      }
    } elsif (/^##\s*MAF\s+BEGIN/oi) {
      my $maf = "";
      while (<>) {
	last if /^##\s*MAF\s+END/oi;
	$maf .= $_;
      }
      $dagmax = 0;
      %mafstruct = ();
      $mafcurrent = 0;
      sentence_start($1);
      ## Handle $maf
      maf_handle($maf);
      sentence_end($dagmax);
      $dagmax = 0;
      %mafstruct = ();
    } elsif (/^<\?xml/oi) {
      my $maf = $_;
      while (<>) {
	$maf .= $_;
	last if m{^</maf>}oi;
      }
      $dagmax = 0;
      %mafstruct = ();
      $mafcurrent = 0;
      sentence_start($1);
      ## Handle $maf
      maf_handle($maf);
      sentence_end($dagmax);
      $dagmax = 0;
      %mafstruct = ();
    } else {
      open2(*OUT,*IN,$cmd);
      print IN $_;
      close(IN);
      line_handler(\"^##DAG BEGIN\n");
      while (<OUT>) {
	line_handler(\$_);
      }
      line_handler(\"^##DAG END\n");
      close OUT;
    }
  }
} else {
  while (<>) {
    line_handler(\$_);
  }
}

sub line_handler {
  my $lref = shift;
  if ($$lref =~ /##\s*DAG\s+BEGIN\s*(.*)/) {
    $dagmode ||= 1;
    $dagmax = 0;
    sentence_start($1);
    return;
  };
  if ($dagmode && $$lref =~ /##\s*DAG\s+END\b/) {
    sentence_end($dagmax);
    $dagmax = 0;
    $dagmode = 0;
    return;
  };
  if ($dagmode) {
    dag_decode($$lref);
  } else {
    segmenter($$lref);
  }
}

$h->finish;

sub sentence_start {
  my $sentence = shift || '';
  @fifo = ();
  $nbsentence++;
  if ($expand) {
  print <<EOF;
%% Token database generated by $0
%% Sentence $nbsentence $sentence

EOF
} else {
  print "%% $sentence\n";
}
} 

sub sentence_end {
  my $n = shift;
  while (@fifo) {
    my $word = shift @fifo;
    handle_word($word) unless ($word->{sent});
  }
  print <<EOF if ($expand);

'N'($n).

EOF

  print <<EOF;
%% SENTENCE DIV %%

EOF

}

sub handle_date {
  my $date = shift;
  my $lemma = quote(shift);
  my $left = shift;
  my $right = shift;
  my $truelex = quote(shift || $date);
  $date = quote($date);
  print <<EOF if ($expand);
'C'($left,
    lemma{ lex => $date,
           truelex => $truelex,
           cat => nc,
           top => nc{gender => masc, number => sg, time => (+)},
           lemma => $lemma
         },
    $right
   ).

EOF
}

sub handle_number {
  my $xnumber = shift;
  my $number = quote($xnumber);
  my $lemma = quote(shift);
  my $left = shift;
  my $right = shift;
  my $gender = shift || '_';
  my $truelex = shift || $xnumber ;
  send_word('_NUMBER',$truelex,"pro|nc",$left,$right,'keep',undef);
  $truelex = quote($truelex);
  print <<EOF if ($expand);
'C'($left,
    lemma{ lex => $number,
           truelex => $truelex,
           cat => number,
           top => number{ gender => $gender },
           lemma => $lemma
         },
    $right
   ).

EOF
}

sub handle_ordinal {
  my $lex = shift;
  my $lemma = quote(shift);
  my $left = shift;
  my $right = shift;
  my $number = shift || '_';
  my $truelex = quote(shift || $lex);
  $lex = quote($lex);
  print <<EOF if ($expand);
'C'($left,
    lemma{ lex => $lex,
           lex => $truelex,
           cat => adj,
           top => adj{ number => $number },
           lemma => $lemma
         },
    $right
   ).

EOF
}


sub handle_word {
  my $data= shift;
  my $lex = $data->{lex};
  my $truelex = $data->{truelex};
  if (!exists $cachelexed{$lex}) {
    $lexin = "$lex\n";
    ## print "SEND INFO TO LEXED $lexin\n";
    $h->pump until $lexout;
    ##    print "GOT INFO FROM LEXED $lexout\n";
    my $info = $lexout;
    chomp $info;
    $cachelexed{$lex} = $info;
    $lexout = '';
  }
  my $info = $cachelexed{$lex};
  my $cat = $data->{cat};
  my %cat = ();
  if (defined $cat) {
    $cat{$_} ||= 1 foreach (split(/\|/,$cat));
  }
  my $left = $data->{left};
  my $right = $data->{right};
  my $status = $data->{status};
  my $altbox = $data->{altbox};
##  print "%% lexed handling '$lex' '$truelex' cat='$cat' status=$status: $info\n";
  if ($info =~ /<unknown>/ ) {
    return if ($status eq 'discard');
    return if (defined $altbox  && $altbox->{found});
    return unless ($expand);
    unless (defined $cat) {
      if ($lex eq '_Uw' || ($lex =~ /^[[:upper:]]/ && $lex !~ /\b-[[:lower:]]/)) {
	$cat = 'np';
      } else {
	$cat = 'cat[v,nc,adv,adj]';
	$unknown{$lex} = 1;
      }
    }
    $lex = quote($lex);
    $truelex = quote($truelex);
    print <<EOF;
'C'($left, 
    lemma{ lex => $lex, truelex => $truelex, cat => $cat },
    $right
   ).
EOF
    return;
  }
  return if (defined $altbox && $altbox->{found} == 1 && $data->{status} eq 'discard_but');
  if ($expand) {
    my $entries = entry_parse($info,$data->{fs});
    foreach my $entry (@$entries) {

      ## print "TEST  entry lemma=$entry->{xlemma} cat=$entry->{cat}\n";
      ## print "  with data lemma=$data->{entry}\n";

      ##      my $lemma = ($entry->{lemma} eq '_' ? $lex : unquote($entry->{lemma}));
      my $lemma = $entry->{xlemma} || $lex;

      next if (defined $cat && !exists $cat{$entry->{cat}});
      next if (exists $data->{entry} && $data->{entry} ne $lemma);
      $altbox->{found} ||= 1 if (defined $altbox);
      entry_print($entry,$truelex,$left,$right,$lex);
    }
  } else {
    print "$data->{truelex}\n";
  }
}

sub entry_parse {
  my $entries = shift;
  my $fs = shift;
##  print "Entry parse: $entries\n";
  my $res  = $parser->entries($entries);
##  print Data::Dumper->Dump([$res],['entry']);
  return [ map($_->lexLP($fs),@{$res}) ];
}

END {
  if (%unknown) {
    my $file = $config->tmpdir()."/frmg.unknown";
    if (-r $file) {
      open(UNKNOWN,"<$file") || die "can't open $file: $!";
      foreach my $word (<UNKNOWN>) {
	chomp $word;
	delete $unknown{$word};
      }
      close(UNKNOWN);
    }
    open(UNKNOWN,">>$file") || die "can't open $file: $!";
    print UNKNOWN "$_\n" foreach (sort keys %unknown);
    close(UNKNOWN);
  }
}

sub segmenter {
  my $line = shift;
  my $i = 0;

  sentence_start($line);

  $line =~ s/([.]+)(?![:\w])/ $1/og;
  $line =~ s/([!?]+)(?!:)/ $1/og;
  $line =~ s/([\(\]+)/$1 /og;
  $line =~ s/([,;?\\)])(?!:)/ $1/og;
  $line =~ s/"/ " /og;
  $line =~ s/:\s+/ : /og;
##  $line =~ s/\bl'on\b/on/ogi;
  $line =~ s/((?:(?:^|\s)\w|qu)\')/$1 /ogi;
  $line =~ s/\b-t-(il|elle|on|y|ils|elles)\b/ -t-$1/ogi;
  $line =~  s/\B([td])-(il|elle|on|y|ils|elles)\b/$1 -$2/ogi;
  $line =~ s/\B-je\b/e -je/ogi;
  $line =~ s/(?<!-t)-(je|tu|nous|vous|ce|ci|mme|le|la|lui|moi|toi|leur|on|en)
	     (?=$|\s+|-(?:le|la|lui|moi|toi|leur|on|en))
	     / -$1/ogix;
  $line =~ s/\b-(ci|l)\b/ -$1/ogi;

 ## $line =~ s/\b($agglpat)\b/$agglutinates{lc($1)}/oig;
 ##  $line =~ s/\b($agglpat)-/$agglutinates{lc($1)}-/oig;

  ##      print STDERR $line;
  ##      $line =~ //g;
  ##  print "%% LINE '$line'\n";
  
  my $origlex = "";

 LOOP: {
    
##    print "%% HERE $i\n";

    redo LOOP if ($line =~ /\G\s+/gc);
    
    last LOOP if ($line =~ /\G\#/gc || $line =~ /\G$/gc);
    my $left = $i;
    my $right = ++$i;

    if ($line =~ /\G$lexinfo/gc) {
      $origlex = clean_comment($1);
##      $origlex =~ s/^(\S+?):/$1\t/o;
##      print "ORIGLEX $origlex\n";
      --$i;
      redo LOOP;
    }


    ## Titles
    if ($line =~ /\G($RE->{title})/gc) {
      my ($title,@comment) = extract_comment($1);
      my $truelex = $origlex ? join(' ',$origlex,@comment) : $title;
      $right += scalar(my @tmp = split(/\s+/,$title)) - 1;
      $i = $right;
      $title =~ s/\s+\./\./og;
      send_word($title,$truelex,'nc',$left,$right,'keep');
      redo LOOP;
    }

    ## Abbreviations
    if ($line =~ /\G($RE->{abbrev})/gc) {
      ##      print "FOUND ABBREV $1\n";
      my ($abbrev,@comment) = extract_comment($1);
      my $truelex = $origlex ? join(' ',$origlex,@comment) : $abbrev;
      $i = $right += scalar(my @tmp = split(/\s+/,$abbrev)) - 1;
      $abbrev =~ s/\s+\././og;
      send_word($abbrev,$truelex,'np',$left,$right,'keep');
      redo LOOP;
    }

    ## Tel
    if ($line =~ /\G($RE->{tel})/gc) {
      ##      print "FOUND TEL $1\n";
      my ($tel,@comment) = extract_comment($1);
      my $truelex = $origlex ? join(' ',$origlex,@comment) : $tel;
      $i = $right += scalar(my @tmp =  split(/\s+/,$tel)) - 1;
      $tel =~ s/\s+\./\./og;
      send_word('_TEL',$truelex,undef,$left,$right,'keep');
      redo LOOP;
    }

    ## Email
    if ($line =~ /\G($RE->{email})\b/gc) {
      ##      print "FOUND EMAIL $1\n";
      my ($email,@comment) = extract_comment($1);
      my $truelex = $origlex ? join(' ',$origlex,@comment) : $email;
      $i = $right += scalar(my @tmp = split(/\s+/,$email)) - 1;
      $email =~ s/\s+\./\./og;
      send_word('_EMAIL',$truelex,undef,$left,$right,'keep');
      redo LOOP;
    }



    ## Roman Number
    if ($line =~ /\G([IVXLCDM]{2,}(:?e|eme(:?s)))(\s+|$)/gc) {
      my $ordinal = $1;
      my $truelex = $origlex || $ordinal;
      my $number = ($ordinal =~ /s$/) ? 'pl' : 'sg';
      handle_ordinal($ordinal,'ROMAN_NUMBER',$left,$right,$number,$truelex);
      redo LOOP;
    }
   
    ## Dates
    if ($line =~ /\G($RE->{date})\b/gc) {
      my ($date,@comment) = extract_comment($1);
      my $truelex = $origlex ? join(' ',$origlex,@comment) :  $date;
      $i = $right += scalar(my @tmp = split(/\s+/,$date)) - 1;
      handle_date($1,'DATE',$left,$right,$truelex);
      redo LOOP;
    }
    
    ## Ordinals
    if ($line =~ /\G($RE->{ordinal})\b/gc) {
      my ($ordinal,@comment) = extract_comment($1);
      my $truelex = $origlex ? join(' ',$origlex,@comment) : $ordinal;
      $i = $right += scalar(my @tmp = split(/\s+/,$ordinal)) - 1;
      my $number = ($ordinal =~ /s$/) ? 'pl' : 'sg';
      handle_ordinal($ordinal,'ORDINAL',$left,$right,$number,$truelex);
      redo LOOP;
    }

    ## Other numbers
    if ($line =~ /\G($RE->{number})\b/gc) {
      my ($number,@comment) = extract_comment($1);
      my $truelex = $origlex ? join(' ',$origlex,@comment) : $number;
      $i = $right += scalar(my @tmp = split(/\s+/,$number)) - 1;
      my $gender = ($number =~ /une$/) ? 'fem' : '_';
      handle_number($number,'_NUMBER',$left,$right,$gender,$truelex);
      redo LOOP;
    }

    if ($line =~ /\G:\s+/gc) {
      my $truelex = $origlex || ':';
      send_word(':',$truelex,undef,$left,$right,'keep',undef);
      redo LOOP;
    }

    ## agglutinates
    if ($line =~ /\G((?:\S+_)?)($agglpat)\b/gc) {
        my $pref = $1;
        my $word = $2;
	my $comment = $origlex ? "$origlex" : "";
        my @words = split(/\s+/,$agglutinates{lc($word)});
	$words[0] = "$pref$words[0]";
        $i = $right += @words - 1;
	my $altbox ||= { found => 0 };
	cut_word($comment,"$pref$word",$left,$right,$altbox,'discard');
	for (my $j = $left; @words ; $j++) {
	  my $word = shift @words;
	  cut_word($comment,$word,$j,$j+1,$altbox);
	  $comment =~ s/(\S+?)\|(.*)/$1\|*/o;
	}
        redo LOOP;
    }

    ## standard tokens
    if ($line =~ /\G(\S+)/gc){
      cut_word($origlex,$1,$left,$right);
      redo LOOP;
    }
  }

  sentence_end($i);

}

sub extract_comment {
  my $s = shift;
  my @s = split(/$lexinfo/,$s);
  my @w = ();
  my @comment = ();
  while (@s) {
    push(@w,shift @s);
    push(@comment,clean_comment(shift @s)) if (@s);
  }
  return (join(' ',@w),@comment);
}

sub clean_comment {
  my $s = shift;
  $s =~ s/\s+//og;
  return $s;
}

sub cut_word {
  my ($truelex,$xlex,$left,$right,$altbox,$status) = @_;
  my ($lex,$cat,$ht,$top) = split(':',$xlex);
##  print "LEX $lex $truelex\n";
  $truelex ||= $lex;
##  $lex =~ s/(?<!_)_(?!_)/ /o;
  $lex =~  s/([^_-])_([^_-])/$1 $2/og;
  $lex =~ s/'\s+/'/o;
  unless ($status) {
    $status = 'keep';
    $status = 'discard' if ($lex =~ m{/\w+/});
  }
  if ($left == 0 && $lex =~ /^[[:upper:]]/) {
    my $llex = lc($lex); 
    $altbox ||= { found => 0 };
    if ($llex eq 'a') {
      send_word('',$truelex,$cat,$left,$right,'discard',$altbox);
    }
    ##        my $lcat = $cat || "det|pri|prep|adv";
    send_word($llex,$truelex,$cat,$left,$right,'discard',$altbox);
    send_word($lex,$truelex,$cat,$left,$right,'keep',$altbox);
  } else {
    send_word($lex,$truelex,$cat,$left,$right,$status,$altbox);
  }
}

sub send_word {
  my ($lex,$truelex,$cat,$left,$right,$status,$altbox) = @_;
  ##  print "%% preparing sending $lex: @fifo\n";
  my @first = ($#fifo > 5 ? @fifo[0 .. 4] : @fifo);
  foreach my $word (@first) {
    ##    print "%% trying merge\n";
    if ($word->{right} == $left) {
      my $newlex = "$word->{lex} $lex";
      $newlex =~ s/'\s+/'/og;
      if (exists $locutions{$newlex} || exists $locutions{lc($newlex)}) {
	my $newtruelex = "$word->{truelex} $truelex";
	$newtruelex =~ s/' /'/og;
	send_word($newlex,$newtruelex,undef,$word->{left},$right,'discard',$word->{altbox});
      } elsif (!$word->{sent}) {
	$word->{sent} = 1;
	handle_word($word);
      }
    }
  }
##  print "%% sending $lex ($truelex)\n";
  my $locution = (exists $locutions{$lex}) || (exists $locutions{lc($lex)});
  my $data = {
	      lex => $lex,
	      truelex => $truelex,
	      cat => $cat,
	      left => $left,
	      right => $right,
	      status => $status,
	      altbox => $altbox,
	      sent => 0
	     };
  if ($locution) {
    ## we wait until next word before deciding trying $lex
    $data->{altbox} = $altbox ||= {found => 0 };
##    $data->{status} = 'discard';
    @fifo = waiting_word($data,@fifo);
    return;
  }
  if ($lex =~ /^(?:$prefpat)([^-]+)/) {
    my $seed = $1;
##    print "Try compound $lex seed=$seed\n";
    $data->{altbox} = $altbox ||= {found => 0};
    $data->{status} = 'discard';
    handle_word($data);
    my $seeddata = {
		    lex => $seed,
		    truelex => $truelex,
		    cat => $cat,
		    left => $left,
		    right => $right,
		    status => 'discard_but',
		    altbox => $altbox,
		    sent => 0
		   };
    handle_word($seeddata);
  } else {
    handle_word($data);
  }
}

sub waiting_word {
  my ($word,@l) = @_;
  if (@l && $word->{left} == $l[0]->{left} &&  $word->{right} == $l[0]->{right}) {
    my $x = shift @l;
    return ($x,waiting_word($word,@l));
  } else {
    return ($word,@l);
  }
}

sub read_macrosfile {
  my $macrosfile = shift;
  open(MACRO,"<$macrosfile") || die "can't open $macrosfile";
  while (<MACRO>) {
    next unless /^\@/;
    my $line = $_;
    $line =~ s{(/{2}.*)$}{};
    chomp $line;
    my ($name,$body) = split(/\s*=\s*/,$line,2);
    next unless ($body =~ /^\[/);
    ##    print "PARSING MACRO $name $body\n";
    ##    my $fs = $parser->macrodef(\$body);
    ##    print dump($fs->{f});
    ##    $macro{$name} = [ %{$fs->{f}} ];
    $macro{$name} = $body;
  }
  close(MACRO);
  return %macro;
}

sub macro_expand ($) {
  my $macro = shift;
  return [] unless exists $macro{$macro};
  my $fs = $parser->macrodef($macro{$macro});
##  print "expand macro $macro: $macro{$macro} -> $fs\n";
  return [ %{$fs->{f}} ];
}

sub quote {
  my $string = shift || "";
  return $string if ($string =~ /^\d+$/);
  return $string if ($string =~ /^[a-z]\w*$/o);
  #    $string =~ s/\'/\'\'/og;
  $string =~ s/\'(?!\')/\'\'/og;
  return "\'$string\'";
}

sub unquote {
  my $string = shift || "";
  $string =~ s/^'(.*)'$/$1/o;
  return $string;
}

sub parse_map_file {
    my $map = shift;
    my $file = shift;
    open("MAP","<$file") || die "can't open map file $file";
    while (<MAP>) {
        chop;
        next if (/^\#/);
        next if (/^\s*$/);
        $map->{$1}{$2} = $3, next if (/^(\w+)\s+(\S+)\s*:\s*(.+)\s*$/);
        $map->{$1} = $2, next if (/^(\w+)\s*:\s*(.+)\s*$/);
    }
    close(MAP);
}

sub dag_decode {
  my $edge = shift;
  my ($left,$comment,$token,$right) = ($edge =~ /^(\d+)\s+\{(.*?)\}\s+(\S+)\s+(\d+)/);

  if ($comment =~ /<F/) {
    ## we are in a DAG generated by sxpipe: beware of positions that starts at 1
    ## => need to decrement because DyALog likes to have DAG pos starting at 0
    $left--;
    $right--;
    my @comments = ();
    while ($comment =~ m{<F\s+id="(.+?)"\s*>\s*(.*?)\s*</F>}og) {
      my ($id,$w) = ($1,$2);
      my $idw = ($config->comment) ? "$id|$w" : "$w";
      push(@comments,$idw) unless grep( $_ eq $idw,@comments);
    }
    $comment = join(' ',@comments);
  }
  my $cat;
  if ($token =~ /^(\S+)__(\S+?)$/) {
      $token = $1;
      $cat = $2;
  }
  $token =~ s/([^_-])_([^_-])/$1 $2/og;
  my $data = { lex => $token,
	       truelex => $comment,
	       left => $left,
	       right => $right,
	       status => 'keep'
	     };
##  print Data::Dumper->Dump([$data],['data']);
  $data->{cat} = $cat if (defined $cat);
  $dagmax = $right if ($right > $dagmax);
  if (@fifo && $fifo[0]->{left} == $left) {
##    unshift @fifo;
    shift @fifo;
  }
  if ($left == 0 && $token eq '_Uw') {
    push(@fifo,$data);
  } else {
    handle_word($data);
  }
}

## MAF Handling

sub maf_handle {
  my $maf = shift;
  my $twig = XML::Twig->new(
			    twig_handlers => {
					      token => \&handle_maf_token,
					      transition => \&handle_maf_transition,
					      'maf/wordForm' => \&handle_maf_wordform,
					      'maf/wfAlt' => \&handle_maf_wfAlt
					     },
			    pretty_print    => 'indented',
			    output_encoding => 'ISO-8859-1',
			    keep_encoding => 1
			   );
  $twig->parse($maf);
  $twig->purge;
}

sub handle_maf_token {
  my ($t,$token) = @_;
##  print "MAF TOKEN ".$token->att("id")." ".($token->att("value") || $token->trimmed_text)."\n";
  $mafstruct{tokens}{$token->att("id")} = $token->att("value") || $token->trimmed_text;
}

sub normalize_lingua_tagvalue {
  my ($f,$v) = @_;
  $f = $lingua_feature_transfer{$f} || $f;
  if ($lingua_value_expand{$f}) {
    my $type = Lingua::Features::FeatureType->type($f);
    $v = $type->value_name($v) || $v;
  }
  return normalize_value($v);
}

sub handle_maf_transition {
  my ($t,$trans) = @_;
  my $left = $trans->att("source");
  my $right = $trans->att("target");
  my $wf = $trans->first_child("wfAlt");
  $trans = $wf if (defined $wf);
  foreach my $wf ($trans->children("wordForm")) {
    handle_maf_wordform($t,$wf,$left,$right);
  }
}

sub handle_maf_wfAlt {
  my ($t,$alt) = @_;
  my $save = $mafcurrent;
  foreach my $wf ($alt->children("wordForm")) {
    handle_maf_wordform($t,$wf);
    $mafcurrent = $save;
  }
  $mafcurrent=$save+1;
}

sub handle_maf_wordform {
  my ($t,$wf,$left,$right) = @_;
##  $left =~ s/[^\d]+//og;
##  $right =~ s/[^\d]+//og;
##  print "MAF left=$left right=$right current=$mafcurrent\n";
  if (defined $left) {
##    $left = $mafstruct{state}{$left} ||= $mafcurrent;
##    $right = $mafstruct{state}{$right} ||=  $left+1;
    $left =~ s/[^\d]+//og;
    $right =~ s/[^\d]+//og;
  } else {
    $left = $mafcurrent;
    $right = $left+1;
  }
  $mafcurrent = $right;
  my @tokens = split(/\s+/,$wf->att("tokens"));
  my $entry = $wf->att("entry");
  $entry =~ s/^urn:lefff://;
  $entry =~ s/_\d+$//;
##  $entry =~ s/ /_/og;
  my $truetoken = join(" ",map($mafstruct{tokens}{$_},@tokens));
  my $token = $wf->att("form") || $truetoken;
  $token =~ s/([^_-])_([^_-])/$1 $2/og;
##  print "MAF TRANS $left $right @tokens $token $entry\n";
  my $fs = {};
  ##  print "Tokens @tokens => $token\n";
  if (defined $wf->child("fs")) {
    my $xmlfs= $wf->child("fs");
    foreach my $f ($xmlfs->children("f")) {
      my $k = $f->att("name");
      my @v = ();
      ##      print "Found ".$f->gi." $k\n";
      foreach ($f->descendants("symbol")) {
	my $value = $_->att("value");
	next unless ($value);
	push(@v,normalize_value($value));
      }
      foreach ($f->descendants("numeric")) {
	my $value = $_->att("value");
	next unless ($value);
	push(@v,normalize_value($value));
      }
      foreach ($f->descendants("binary")) {
	my $value = $_->att("value");
	next unless ($value);
	push(@v,normalize_value($value));
      }
      ## print "Found values @v\n";
      $fs->{normalize_feature($k)} = [@v];
    }
  } elsif (defined $wf->att("tag")) {
    my $tag = $wf->att("tag");
##    $tag =~ s/\bpos\./cat./og;
##    print "MAF $tag\n";
    foreach my $feature (split(/\s+/,$tag)) {
      my ($f,@v) = split(/\./,$feature);
##      print "MAF TYPE $f '$type'\n";
##      @v = map( normalize_value( ($type && $type->value_name($_)) || $_), @v);
      @v = map( normalize_lingua_tagvalue($f,$_), @v);
##      print "MAF2 $f @v\n";
      $f = normalize_feature($f);
      $fs->{$f} = [ @v ];
    }
  }
  my $data = { lex => $token,
	       truelex => $truetoken,
	       left => $left,
	       right => $right,
	       status => 'keep',
	       cat => join('|',@{$fs->{pos}}),
	       entry => $entry,
	       fs => $fs
	     };
##  print Data::Dumper->Dump([$data],['data']);
  $dagmax = $right if ($right > $dagmax);
#  print "Ready to handle word '$entry' @tokens '$token' $left $right '$data->{cat}'\n";
  if (@fifo && $fifo[0]->{left} == $left) {
#    unshift @fifo;
    shift @fifo;
  }
  if ($left == 0 && $token eq '_Uw') {
    push(@fifo,$data);
  } else {
    handle_word($data);
  }
}

sub build_elt {
  my $v = shift;
  if ($v eq '+') {
    return TAG::Plus->new;
  } elsif ($v eq '-') {
    return TAG::Minus->new;
  } else {
    return TAG::Val->new({text=>$v});
  }
}

sub fs_unify {
  my ($fs1,$fs2) = @_;
  ## print "TRY UNIFY\n";
  ## Warning: $fs1 and $fs2 are both fs but do not follow the same data model
  foreach my $k (keys %{$fs2}) {
    next if ($k =~ /pos/ || $k =~ /fct/ || $exclude_feature{$k});
    if (exists $fs1->{f}{$k}) {
      ##    print "CHECK KEY $k\n";
      my $v2 = $fs2->{$k};
      my $v1 = $fs1->{f}{$k};
      foreach my $v (@$v2) {
	my $vv = build_elt($v);
	## print "Check compare on $k:$v => $vv\n";
	##  print "\twith ".join(' ',map($_->text,@{$v1->{value}}))."\n";
	return 0 unless grep(Compare($vv,$_),@{$v1->{value}});
	## print "Check compare on $k:$v => OK\n";
      }
    } else {
      ##      $clone->();
      ## print "ADD KEY $k\n";
      $fs1->{f}{$k} = TAG::Comp->new( { name => $k,
					rel => 'eq',
					value => [ map( build_elt($_), @{$fs2->{$k}} ) ]
				      } );
    }
  }
  return 1;
}

1;

__END__
